#include "stdlib.h"
#include "string.h"
#include "genarray.h"

/** implementation of the data structure **/

struct _genarray {
  size_t ga_size;		// Current number of elements.
  size_t ga_element_size; 
  void *ga_mem;			// Actual array.
  void *ga_def;			// Default value.
};

genarray *ga_new(size_t initial_size, size_t element_size, const void *default_value) {
  genarray *r = malloc (sizeof (struct _genarray));
  if (r != NULL) {
    r->ga_size = initial_size;
    r->ga_element_size = element_size;
    r->ga_def = malloc (element_size);
    r->ga_mem = calloc (initial_size, element_size);
    if (r->ga_def != NULL && r->ga_mem != NULL) {
      memcpy (r->ga_def, default_value, element_size);
      for (int i=0; i<initial_size; i++) {
	memcpy (r->ga_mem + i*element_size, default_value, element_size);
      }
    } else {
      if (r->ga_def != NULL) free (r->ga_def);
      if (r->ga_mem != NULL) free (r->ga_mem);
      return NULL;
    }
  }
  return r;
}

void ga_destroy (genarray *ga) {
  if (ga != NULL) {
    free (ga->ga_def);		/* free default value */
    free (ga->ga_mem); 		/* free interal array */
    free (ga);			/* free the struct */
  }
}

void ga_read(const genarray *ga, size_t i, void *return_val) {
  if (ga == NULL) {
    return ;
  } else { // ga != NULL
    if (i < ga->ga_size) { // index i is valid!
      memcpy (return_val, ga->ga_mem + i*ga->ga_element_size, ga->ga_element_size);
      return;
    } else { 			/* index exceeds array */
      memcpy (ga->ga_def, return_val, ga->ga_element_size);
      return;
    }
  }
}

int ga_write (genarray *ga, size_t i, const void *val) {
  if (ga == NULL) {
    return -1;
  } else {
    if (i < ga->ga_size) { 	/* index i is valid */
      memcpy (ga->ga_mem + i*ga->ga_element_size, val, ga->ga_element_size);
      return ga->ga_size;
    } else { 			/* index exceeds array */
      int newsize = i+1;	/* so that i is a valid index */
      ga->ga_mem = realloc (ga->ga_mem, newsize * ga->ga_element_size);
      if (ga->ga_mem == NULL) {
	return -1;
      }
      // initialize the new part of the array
      for (int j=ga->ga_size; j<i; j++){
	memcpy (ga->ga_mem + j*ga->ga_element_size, ga->ga_def, ga->ga_element_size);
      }
      memcpy (ga->ga_mem + i*ga->ga_element_size, val, ga->ga_element_size);
      ga->ga_size = newsize;
      return newsize;
    }
  }
}
