#include <unistd.h>
#include "time.h"
#include "stdlib.h"

#include "./tui.h"
#include "genarray.h"

#define NORTH 'i'
#define SOUTH 'm'
#define EAST 'k'
#define WEST 'j'

int distance (Size2 * p, Size2 * q) {
  return (p->x-q->x)*(p->x-q->x) + (p->y-q->y)*(p->y-q->y);
}

int main(void) {
  tui_init();

  char wormh = 'o';
  char wormt = 'x';
  size_t t = 0;

  Size2 wsize = tui_size();	/* don't resize window! */
  Size2 size = {.x = wsize.x, .y = wsize.y - 1};
  Size2 worm_head_pos = {.x = size.x/2, .y = size.y/2};
  int worm_feed = 10;		/* worm starts by growing to length 10 */
  int worm_length = 1;
  bool worm_crash = false;
  int worm_direction = EAST;	/* worm starts going east */
  int worm_speed = 50;		/* medium speed */

  genarray *the_worm = ga_new (1, sizeof (Size2), &worm_head_pos);

#define NGOODIES 3
#define GOODY_NULL 0
#define GOODY_CASH '$'
#define GOODY_GEM  '!'
#define GOODY_FOOD '%'
  int goody_kind[NGOODIES] = { GOODY_FOOD, GOODY_CASH, GOODY_GEM };
  int goody_low[NGOODIES]  = { 1, 5, 10 };
  int goody_span[NGOODIES] = { 4, 5, 10 };
  int goody_speedup[NGOODIES] = { 0, 1, 4 };
  int distribution[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2 };

#define NDISTRIBUTION (sizeof (distribution) / sizeof (int))

  typedef struct _goody {
    int kind;
    int value;
    int speedup;
    Size2 position;
  } goody;

  /* seed the random generator */
  time_t tm;
  srand ((unsigned) time (&tm));

#define NGOODIES_IN_GAME 100
  goody goodies[NGOODIES_IN_GAME] = { 0 };
  for (size_t i = 0; i < NGOODIES_IN_GAME; i++) {
    int d = distribution[rand() % NDISTRIBUTION];
    goodies[i].kind = goody_kind[d];
    goodies[i].value = goody_low[d] + rand () % goody_span[d];
    goodies[i].speedup = goody_speedup[d] ? goody_speedup[d] + (rand() % goody_speedup[d]) : 0;
    do {
      goodies[i].position.x = rand () % size.x;
      goodies[i].position.y = rand () % size.y;
    } while (distance (&goodies[i].position, &worm_head_pos) < 25);
  }
  
  while (1) {
    if (worm_crash) {
      /* you lose */
      while (read_from_stdin() != 'q');
      break;
    }
    if (stdin_has_changed()) {
      char c = read_from_stdin();
      if (c == 'q')
        break;
      switch (c) {
      case NORTH:
      case SOUTH:
      case EAST:
      case WEST:
	worm_direction = c;
	break;
      }
    }
    /* step the model */
    switch (worm_direction) {
    case NORTH:
      if (worm_head_pos.y > 0) {
	worm_head_pos.y--;
      } else {
	worm_crash = true;
      }
      break;
    case SOUTH:
      if (worm_head_pos.y < size.y - 1) {
	worm_head_pos.y++;
      } else {
	worm_crash = true;
      }
      break;
    case WEST:
      if (worm_head_pos.x > 0) {
	worm_head_pos.x --;
      } else {
	worm_crash = true;
      }
      break;
    case EAST:
      if (worm_head_pos.x < size.x - 1) {
	worm_head_pos.x ++;
      } else {
	worm_crash = true;
      }
    }
    /* grow the worm, if there is left over food */
    if (worm_feed > 0) {
      ga_write (the_worm, worm_length, &worm_head_pos);
      worm_length ++;
      worm_feed --;
    } else {
      for (int j = 0; j < worm_length - 1; j ++) {
	Size2 worm_tail;
	ga_read (the_worm, j+1, &worm_tail);
	ga_write (the_worm, j, &worm_tail);
      }
      ga_write (the_worm, worm_length - 1, &worm_head_pos);
    }
    
    /* what if position of worm head is already occupied? */
    for (int j = 0; j < worm_length - 1; j++) {
      Size2 worm_tail;
      ga_read (the_worm, j, &worm_tail);
      worm_crash = (worm_tail.x == worm_head_pos.x && worm_tail.y == worm_head_pos.y);
      if (worm_crash) {
	break;
      }
    }
    /* check for goodies */
    if (!worm_crash) {
      for (int j = 0; j < NGOODIES_IN_GAME; j++) {
	if (goodies[j].kind && goodies[j].position.x == worm_head_pos.x && goodies[j].position.y == worm_head_pos.y) {
	  goodies[j].kind = GOODY_NULL;
	  worm_feed += goodies[j].value;
	  if (worm_speed > 10) {
	    worm_speed -= goodies[j].speedup;
	  }
	}
      }
    }
    /* model to view */
    tui_clear();

    Cell* cell = tui_cell_at(worm_head_pos.x, worm_head_pos.y);
    cell->content = wormh;
    cell->text_color = FG_HI_GREEN;
    cell->background_color = BG_BLACK;
    for (int j = worm_length - 2; j > 0; j--) {
      Size2 worm_tail;
      ga_read (the_worm, j, &worm_tail);
      cell = tui_cell_at(worm_tail.x, worm_tail.y);
      cell->content = wormt;
      cell->text_color = FG_GREEN;
      cell->background_color = BG_BLACK;
    }
    for (int j = 0; j < NGOODIES_IN_GAME; j++) {
      if (goodies[j].kind) {
	cell = tui_cell_at (goodies[j].position.x, goodies[j].position.y);
	cell->content = goodies[j].kind;
	cell->text_color = FG_HI_RED;
	cell->background_color = BG_BLACK;
      }
    }

    char buf[255];
    sprintf(buf, "LENGTH: %4d    FEED: %2d    SPEED: %2d    %s",
            worm_length, worm_feed, 60 - worm_speed, worm_crash ? "CRASHED": "");
    tui_set_str_at(0, size.y, buf, FG_WHITE, BG_BLACK);

    tui_update();

    t += 1;

    usleep(worm_speed * 10000);
  }

  /* cleanup */
  ga_destroy (the_worm);

  tui_shutdown();

  return 0;
}
